-- icmp64.vhdl -- F-CPU 64-bit Integer Compare Unit
-- Copyright (C) 2002 Michael Riepe <michael@stud.uni-hannover.de>
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

-- $Id: icmp64.vhdl,v 1.1 2002/07/26 16:58:20 michael Exp $

library IEEE;
use IEEE.std_logic_1164.all;
use work.Bit_Manipulation.all;

entity ICmp64 is
	generic (
		WIDTH : natural := 64	-- do not change!
	);
	port (
		-- operand inputs
		A : in std_ulogic_vector(WIDTH-1 downto 0);
		B : in std_ulogic_vector(WIDTH-1 downto 0);
		-- operating modes
		Comp : in std_ulogic;
		Sort : in std_ulogic;
		Msb0 : in std_ulogic;
		Msb1 : in std_ulogic;
		-- flags
		SignedMode : in std_ulogic;
		-- SIMD mode switches
		U : in std_ulogic_vector(2 downto 0);
		-- clock/reset inputs
		Clk : in std_ulogic;
		Rst : in std_ulogic;
		En : in std_ulogic;
	--
		Yl : out std_ulogic_vector(WIDTH-1 downto 0);
		Yh : out std_ulogic_vector(WIDTH-1 downto 0)
	);
--pragma synthesis_off
begin
	assert WIDTH = 64
		report "width of ICmp must be 64"
		severity failure;
--pragma synthesis_on
end ICmp64;

architecture Behave_1 of ICmp64 is
	signal M : std_ulogic_vector(7 downto 0);
	signal r_M : std_ulogic_vector(7 downto 0);
	signal r_A : std_ulogic_vector(WIDTH-1 downto 0);
	signal r_B : std_ulogic_vector(WIDTH-1 downto 0);
	signal r_C : std_ulogic_vector(WIDTH-1 downto 0);
	signal r_X08 : std_ulogic_vector(WIDTH-1 downto 0);
	signal r_X16 : std_ulogic_vector(WIDTH-1 downto 0);
	signal r_X32 : std_ulogic_vector(WIDTH-1 downto 0);
	signal r_X64 : std_ulogic_vector(WIDTH-1 downto 0);
	signal r_En : std_ulogic;
	signal Ymin : std_ulogic_vector(WIDTH-1 downto 0);
	signal Ymax : std_ulogic_vector(WIDTH-1 downto 0);
	signal Ycmp : std_ulogic_vector(WIDTH-1 downto 0);
	signal Ymsb : std_ulogic_vector(WIDTH-1 downto 0);
begin
	M <= (
		0 => Comp,
		1 => Sort,
		2 => Msb0,
		3 => Msb1,
		4 => SignedMode,
		5 => U(0),
		6 => U(1),
		7 => U(2)
	);

	stage_1 : process (A, B, M, Clk, Rst, En)
		function chunk (X : in std_ulogic_vector;
						N : in natural) return std_ulogic_vector is
			constant w : natural := X'length;
			alias xx : std_ulogic_vector(w-1 downto 0) is X;
			variable yy : std_ulogic_vector(w-1 downto 0);
			variable tt : std_ulogic_vector(N-1 downto 0);
		begin
--pragma synthesis_off
			assert N > 0;
			assert w >= N;
			assert w mod N = 0;
--pragma synthesis_on
			for i in w/N-1 downto 0 loop
				tt := xx(N*i+N-1 downto N*i);
				tt := bit_reverse(cascade_or(bit_reverse(tt)));
				tt := tt xor rshift(tt, 1);
				yy(N*i+N-1 downto N*i) := tt;
			end loop;
			return yy;
		end chunk;

		variable mm : std_ulogic_vector(7 downto 0);
		variable aa : std_ulogic_vector(WIDTH-1 downto 0);
		variable bb : std_ulogic_vector(WIDTH-1 downto 0);
		variable cc : std_ulogic_vector(WIDTH-1 downto 0);
		variable xx : std_ulogic_vector(WIDTH-1 downto 0);
		variable x08 : std_ulogic_vector(WIDTH-1 downto 0);
		variable x16 : std_ulogic_vector(WIDTH-1 downto 0);
		variable x32 : std_ulogic_vector(WIDTH-1 downto 0);
		variable x64 : std_ulogic_vector(WIDTH-1 downto 0);
	begin
		aa := to_X01(A);
		bb := to_X01(B);
		mm := to_X01(M);

		-- tree input select
		-- d=2
		if mm(3) = '1' then
			-- Msb1
			xx := aa;
		elsif mm(2) = '1' then
			-- Msb0
			xx := not aa;
		else
			-- Comp/Sort
			xx := aa xor bb;
		end if;

		-- OR trees
		-- d=6
		x08 := chunk(xx,  8);
		x16 := chunk(xx, 16);
		x32 := chunk(xx, 32);
		x64 := chunk(xx, 64);

		-- handle signed compare/sort
		-- d=3
		cc := (others => '0');
		cc(63) := mm(4);
		cc(55) := mm(4) and not mm(5);
		cc(47) := mm(4) and not mm(6);
		cc(39) := mm(4) and not mm(5);
		cc(31) := mm(4) and not mm(7);
		cc(23) := mm(4) and not mm(5);
		cc(15) := mm(4) and not mm(6);
		cc( 7) := mm(4) and not mm(5);
		cc := cc xor bb;

		-- pipeline register
		if to_X01(Rst) = '1' then
			r_A <= (others => '0');
			r_B <= (others => '0');
			r_C <= (others => '0');
			r_M <= (others => '0');
			r_X08 <= (others => '0');
			r_X16 <= (others => '0');
			r_X32 <= (others => '0');
			r_X64 <= (others => '0');
			r_En <= '0';
		elsif rising_edge(Clk) then
			if to_X01(En) = '1' then
				r_A <= aa;
				r_B <= bb;
				r_C <= cc;
				r_M <= mm;
				r_X08 <= x08;
				r_X16 <= x16;
				r_X32 <= x32;
				r_X64 <= x64;
			end if;
			r_En <= En;
		end if;
	end process;

	stage_2 : process (r_A, r_B, r_C, r_M, r_X08, r_X16, r_X32, r_X64)
		procedure chunk (A, B, C, X : in std_ulogic_vector;
						 YMin, YMax, YCmp, YMsb : out std_ulogic_vector) is
			constant w : natural := X'length;
			alias aa : std_ulogic_vector(w-1 downto 0) is A;
			alias bb : std_ulogic_vector(w-1 downto 0) is B;
			alias cc : std_ulogic_vector(w-1 downto 0) is C;
			alias xx : std_ulogic_vector(w-1 downto 0) is X;
			variable min : std_ulogic_vector(w-1 downto 0);
			variable max : std_ulogic_vector(w-1 downto 0);
			variable cmp : std_ulogic_vector(w-1 downto 0);
			variable msb : std_ulogic_vector(w-1 downto 0);
			variable tt : std_ulogic_vector(w/2-1 downto 0);
			variable less : std_ulogic;
			variable k : natural;
		begin
			less := reduce_or(cc and xx);
			cmp := (others => less);
			if to_X01(less) = '1' then
				min := aa;
				max := bb;
			else
				min := bb;
				max := aa;
			end if;
			msb := (others => '0');
			for i in 0 to 31 loop
				exit when 2**i > w;
				k := 0;
				for j in 1 to w loop
					if j mod 2**(i+1) >= 2**i then
						tt(k) := xx(j - 1);
						k := k + 1;
					end if;
				end loop;
				assert k > 0;
				assert k <= w/2;
				msb(i) := reduce_or(tt(k-1 downto 0));
			end loop;
			YMin := min;
			YMax := max;
			YCmp := cmp;
			YMsb := msb;
		end chunk;

		variable mm : std_ulogic_vector(7 downto 0);
		variable min08, max08, cmp08, msb08 : std_ulogic_vector(WIDTH-1 downto 0);
		variable min16, max16, cmp16, msb16 : std_ulogic_vector(WIDTH-1 downto 0);
		variable min32, max32, cmp32, msb32 : std_ulogic_vector(WIDTH-1 downto 0);
		variable min64, max64, cmp64, msb64 : std_ulogic_vector(WIDTH-1 downto 0);
	begin
		-- postprocessing
		-- d=5
		for i in WIDTH/8-1 downto 0 loop
			chunk(
				r_A( 8*i+ 7 downto  8*i),
				r_B( 8*i+ 7 downto  8*i),
				r_C( 8*i+ 7 downto  8*i),
				r_X08( 8*i+ 7 downto  8*i),
				min08( 8*i+ 7 downto  8*i),
				max08( 8*i+ 7 downto  8*i),
				cmp08( 8*i+ 7 downto  8*i),
				msb08( 8*i+ 7 downto  8*i));
		end loop;
		for i in WIDTH/16-1 downto 0 loop
			chunk(
				r_A(16*i+15 downto 16*i),
				r_B(16*i+15 downto 16*i),
				r_C(16*i+15 downto 16*i),
				r_X16(16*i+15 downto 16*i),
				min16(16*i+15 downto 16*i),
				max16(16*i+15 downto 16*i),
				cmp16(16*i+15 downto 16*i),
				msb16(16*i+15 downto 16*i));
		end loop;
		for i in WIDTH/32-1 downto 0 loop
			chunk(
				r_A(32*i+31 downto 32*i),
				r_B(32*i+31 downto 32*i),
				r_C(32*i+31 downto 32*i),
				r_X32(32*i+31 downto 32*i),
				min32(32*i+31 downto 32*i),
				max32(32*i+31 downto 32*i),
				cmp32(32*i+31 downto 32*i),
				msb32(32*i+31 downto 32*i));
		end loop;
		for i in WIDTH/64-1 downto 0 loop
			chunk(
				r_A(64*i+63 downto 64*i),
				r_B(64*i+63 downto 64*i),
				r_C(64*i+63 downto 64*i),
				r_X64(64*i+63 downto 64*i),
				min64(64*i+63 downto 64*i),
				max64(64*i+63 downto 64*i),
				cmp64(64*i+63 downto 64*i),
				msb64(64*i+63 downto 64*i));
		end loop;

		-- outputs
		-- d=6
		mm := to_X01(r_M);
		if mm(7) = '1' then
			Ymin <= min64;
			Ymax <= max64;
			Ycmp <= cmp64;
			Ymsb <= msb64;
		elsif mm(6) = '1' then
			Ymin <= min32;
			Ymax <= max32;
			Ycmp <= cmp32;
			Ymsb <= msb32;
		elsif mm(5) = '1' then
			Ymin <= min16;
			Ymax <= max16;
			Ycmp <= cmp16;
			Ymsb <= msb16;
		else
			Ymin <= min08;
			Ymax <= max08;
			Ycmp <= cmp08;
			Ymsb <= msb08;
		end if;
	end process;

	out_mux : process (r_M, Ymin, Ymax, Ycmp, Ymsb)
		variable mm : std_ulogic_vector(7 downto 0);
	begin
		-- XXX: I violate the `Six Gate Rule' here.  Since there's
		-- no XOR gate in the datapath of the second stage,
		-- I may get away with it...
		-- d=7
		mm := to_X01(r_M);
		case mm(3 downto 0) is
			when "1000" | "0100" => -- Msb1/Msb0
				Yl <= Ymsb;
				Yh <= (others => 'X');
			when "0010" => -- Sort
				Yl <= Ymin;
				Yh <= Ymax;
			when "0001" => -- Comp
				Yl <= Ycmp;
				Yh <= (others => 'X');
			when others =>
				Yl <= (others => 'X');
				Yh <= (others => 'X');
		end case;
	end process;
end Behave_1;

-- vi: set ts=4 sw=4 equalprg="fmt -72 -p--": please
